/* Terminalhanterare fr *:85 HT2002                     */
/* Av Jozef Swiatycki                                    */

#ifndef _TERMINAL_STAR85_H
#define _TERMINAL_STAR85_H

#include <vector>
#include <iostream>
#include <string>

namespace terminal_star85 {

// Fljande struct Event returneras av terminal::getNextEvent()
// whereX och whereY r kolumn- och radnr
// what r ett vrde av typen terminal::Keycodes motsvarande ASCII-koden 
// fr styr- och skrivbara tecken eller en funktionstangentkod enligt 
// terminal::Keycodes (se nedan)

  struct Event; // Deklaration av Event efter deklarationen av terminal

  class terminal{
  public:
    terminal();             // Initierar terminalfnstret
    ~terminal();            // terstller terminalfnstret
    Event getNextEvent();   // Returnerar Event fr nsta tangenttryckning
    void clrscr();          // Blankstller skrmen
    void clreod();          // Blankstller resten av skrmen frn markren
    void clrbod();          // Blankstller brjan av skrmen till markren
    void clrln();           // Blankstller raden dr markren r
    void clreol();          // Blankstller resten av raden frn markren
    void clrbol();          // Blankstller brjan av raden till markren
    void curmove(int x, int y);  // Flyttar markren till kolumn x, rad y
    void chscroll(int x, int y); // ndrar scrollregionen till x till y
    void curdown(int n=1);  // Flyttar markren n rader ner
    void curleft(int n=1);  // Flyttar markren n positioner till vnster
    void curright(int n=1); // Flyttar markren n positioner till hger
    void curup(int n=1);    // Flyttar markren n rader upp
    void curget(int&, int&);// Ger kolumn, rad dr markren befinner sig
    void inschar(char c);   // Stter in c vid markren
    void insstr(const char *);// Stter in strngen vid markren
    void inslines(int n=1); // Stter in n tomma rader
    void delchar(int n=1);  // Bort tkn hger om markren
    void clrchrs(int n=1);  // Blankstller n tecken fom markren
    void dellines(int n=1); // Tar bort n rader frn aktuell rad
    void scrollup(int n=1); // Scrollar scrollregionen upp n rader
    void scrolldown(int n=1); // Scrollar scrollregionen ner n rader
    void savecur();         // Spara markrpositionen
    void restcur();         // terstll senast sparade markrpositionen
    void bold();            // Efterfljande utskrifter med fet stil
    void reverse();         // Efterfljande utskrifter med omvnd frg
    void underline();       // Efterfljande utskrifter understrukna
    void normal();          // Efterfljande utskrifter med normal stil
    
    enum Colour {Black, Red, Green, Yellow, Blue, Magenta, 
		 Cyan, White, DefColour=9};
    void fgcolour(Colour colr); // Efterfljande text i angiven frg
    void bgcolour(Colour colr); // Efterfljande bakgrund i angiven frg
    
    void horline(int row, int fcol, int tcol); // Ritar horisontell linje
    void verline(int col, int frow, int trow); // Ritar vertikal linje
    void box(int a, int b, int c, int d);      // Ritar en rektangel
    void table(int x, int y, std::vector<int> rows, std::vector<int> cols); 
    // Ritar en tabell: x anger startkolumnen, y anger startraden, ...
    // ..rows anger antalet skrmrader fr varje tabellrad, ...
    // ..cols anger antalet skrmkolumner fr varje tabellkolumn, ...
    // ..bde rows och cols anges exklusive ramen
    
    int height();           // Returnerar terminalfnstrets hjd i rader
    int width();            // Returnerar terminalfnstrets bredd i kolumner
    bool changed();         // Anger om fnstrets storlek har ndrats
                            
    void onresize(void (*func)()); // Installerar func som resize-funktion
    void setsize(int height, int width); // Stter fnsterstorleken
    
    static std::string version();      // Returnerar versionsnummer

    enum Keycodes {  // *** Kontrolltecken
      NULLchar, CTRLspace = 0,
      CTRLa, 
      CTRLb, 
      CTRLc, // Avbryter exekvering, kan ej fngas upp
      CTRLd, 
      CTRLe, 
      CTRLf, 
      CTRLg, BELL = 7,
      CTRLh, BACKSPACE = 8,
      CTRLi, TAB = 9,
      CTRLj, NEWLINE = 10,
      CTRLk, 
      CTRLl, 
      CTRLm, RETURN = 13, ENTER = 13,
      CTRLn, 
      CTRLo, // Kan ej fngas upp
      CTRLp,
      CTRLq, // Kan ej fngas upp
      CTRLr, 
      CTRLs, // Stoppar datafldet, kan ej fngas upp
      CTRLt, 
      CTRLu, 
      CTRLv,
      CTRLw,
      CTRLx,
      CTRLy,
      CTRLz, // Lgger processen i bakgrunden, kan ej fngas upp
      ESC,   // Kan ej fngas upp
      
      DELkey=127, RUBOUT=127,

      // *** Funktionstangenter
      F1=256, 
      F2,	   
      F3,     
      F4,     
      F5,     
      F6, 
      F7, 
      F8, 
      F9, 
      F10, 
      F11,
      F12,
      F13, 
      F14,
      Helpkey, 
      Executekey, 
      F17, 
      F18, 
      F19, 
      F20,
      Homekey, 
      Insertkey, 
      Deletekey, 
      Endkey, 
      Pageupkey, 
      Pagedownkey, 
      Uparrowkey, 
      Downarrowkey, 
      Leftarrowkey, 
      Rightarrowkey,
      // Tangenter p Numeric Key Pad
      PF1, 
      PF2, 
      PF3, 
      PF4, 
      NKPmult,
      NKPplus,
      NKPdiv,
      NKP0, 
      NKP1, 
      NKP2, 
      NKP3, 
      NKP4, 
      NKP5, 
      NKP6, 
      NKP7, 
      NKP8, 
      NKP9, 
      NKPminus, 
      NKPcomma,
      NKPpoint, 
      NKPEnter
    };

  private:
    terminal(const terminal&){}     // Fr att frhindra kopiering
    static bool exists;             // Fr att kolla om ett objektt redan finns
  };
  
  // structen Event som returneras av terminal::getNextEvent()
  struct Event {
    int whereX, whereY;           // Koordinater fr hndelsen (kol, rad)
    terminal::Keycodes what;      // Hndelsens tangentkod
  };
  
  // Implementering av manipulatorer fr terminalen, s att 
  // medlemsfunktioner kan anropas med hjlp av verlagringar av operator<<
  
  struct tmanip2{
    terminal& (*f)(terminal&, int, int);
    int x,y;
    tmanip2(terminal& (*ff)(terminal&,int,int),int xx,int yy):
      f(ff),x(xx),y(yy){}
  };
  
  struct tmanip1{
    terminal& (*f)(terminal&, int);
    int x;
    tmanip1(terminal& (*ff)(terminal&, int),int xx):f(ff),x(xx){}
  };
  
  struct tmanipc{
    terminal& (*f)(terminal&, char);
    char c;
    tmanipc(terminal& (*ff)(terminal&, char), char cc):f(ff),c(cc){}
  };
  
  struct tmanips{
    terminal& (*f)(terminal&, const char*);
    const char *str;
    tmanips(terminal& (*ff)(terminal&, const char *), const char* s):
      f(ff),str(s) {}
  };
  
  tmanip2 curmove(int x, int y);
  tmanip2 chscroll(int x, int y);
  tmanip1 fgcolour(int x);
  tmanip1 bgcolour(int x);
  tmanip1 curright(int n=1);
  tmanip1 curleft(int n=1);
  tmanip1 curdown(int n=1);
  tmanip1 curup(int n=1);
  tmanip1 inslines(int n=1);
  tmanip1 delchar(int n=1);
  tmanip1 clrchrs(int n=1);
  tmanip1 dellines(int n=1);
  tmanip1 scrollup(int n=1);
  tmanip1 scrolldown(int n=1);
  tmanipc inschar(char c);
  tmanips insstr(const char *str);
  terminal& clrscr(terminal& t);
  terminal& clreod(terminal& t);
  terminal& clrbod(terminal& t);
  terminal& clreol(terminal& t);
  terminal& clrbol(terminal& t);
  terminal& clrln(terminal& t);
  terminal& normal(terminal& t);
  terminal& bold(terminal& t);
  terminal& reverse(terminal& t);
  terminal& underline(terminal& t);
  terminal& savecur(terminal& t);
  terminal& restcur(terminal& t);
  
  
  terminal& operator<<(terminal& term, terminal& (*f)(terminal&));
  terminal& operator<<(terminal& term, const tmanip1 m);
  terminal& operator<<(terminal& term, const tmanip2 m);
  terminal& operator<<(terminal& term, const tmanipc m);
  terminal& operator<<(terminal& term, const tmanips m);
  
  template<class TYP>
  terminal& operator<<(terminal& term, const TYP& data){
    std::cout << data << std::flush;
    return term;
  }
  
  terminal& operator<<(terminal& term, const char* data);
  terminal& operator<<(terminal& term, const std::string& data);
  terminal& operator<<(terminal& term, const char data);
  
} // namespace terminal_star85

#endif

  /*********************************************************/
  /* Fljande rader r till fr versionshanteraren RCS     */
  /* Bry er inte om dem                                    */
  /* $Log: terminal.h,v $
   * Revision 2.0  2002/02/11 18:23:16  jozef
   * terminal och hjlpfunktioner nu i namnrymden terminal_star85
   *
   * Revision 1.9  2001/01/26 16:48:38  jozef
   * Anpassat funktionskoderna till PC-tangentbord
   *
   * Revision 1.8  2000/01/09 12:36:49  jozef
   * Anpassat till MS Windows-versionen.
   * Tagit bort #define DTTERM och #ifdef DTTERM
   * Lagt till specialiseringar av terminal& operator<<(terminal&, const TYP&)
   * fr tecken-data fr att mjliggra konverteringar till OEM-teckenkod
   * Lagt till static string terminal::version() fr att kunna avlsa vesionsstmpeln
   *
   * Revision 1.7  1998/12/18 06:32:59  jozef
   * Rttat en kommentar
   *
   * Revision 1.6  1998/12/18 02:41:34  jozef
   * ndrat typen fr Event::what frn int till terminal::Keycodes
   *
   * Revision 1.5  1998/12/17 22:15:16  jozef
   * Anpassat till HT98
   *
   * Revision 1.4  1998/01/02 11:25:36  jozef
   * *** empty log message ***
   *
   * Revision 1.3  1997/12/24 15:54:59  jozef
   * - lagt till prefixet z till anrop av termfunktioner
   *   fr att frhindra namnkollisioner med manipulatorer
   * - lagt till dtterm-funktionalitet
   * - lagt till manipulatorer och verlagringar av operator<<
   * - inkluderat iostream.h och vector
   *
   * Revision 1.1  1997/01/08  04:54:56  jozef
   * Initial revision
   *                                                 */
  /*                 Slut p versionshanterarrader         */
  /*********************************************************/
  
